from datetime import timedelta

from django.db import models
from django.utils import timezone


class Profile(models.Model):

    class Meta:
        db_table = 'instagram_profile'
        ordering = ['username']

    username = models.CharField(max_length=256)
    access_token = models.CharField(max_length=256, blank=True, help_text='The long lived access token')
    expires_at = models.DateTimeField(null=True, blank=True, help_text='The expiry of the stored long lived access token')
    modified = models.DateTimeField(auto_now=True)

    def __str__(self):
        if self.username:
            return '@{}'.format(self.username)
        return 'New profile'

    @property
    def expires_in(self):
        if self.expires_at:
            return self.expires_at - timezone.now()
        return timedelta(seconds=0)

    @property
    def access_valid(self):
        return self.access_token and self.expires_at and self.expires_at > timezone.now()


class Post(models.Model):

    class Meta:
        db_table = 'instagram_posts'
        ordering = ['-created']

    TYPE_IMAGE = 'IMAGE'
    TYPE_VIDEO = 'VIDEO'
    TYPE_ALBUM = 'CAROUSEL_ALBUM'
    MEDIA_TYPES = [
        (TYPE_IMAGE, 'Image'),
        (TYPE_ALBUM, 'Album'),
        (TYPE_VIDEO, 'Video')
    ]

    profile = models.ForeignKey(Profile, null=True, blank=True, on_delete=models.SET_NULL)
    media_id = models.CharField(max_length=100)
    type = models.CharField(max_length=20, choices=MEDIA_TYPES, default=TYPE_IMAGE)
    permalink = models.CharField(max_length=256)
    caption = models.TextField(default='')
    thumbnail = models.ImageField(max_length=256, default='', upload_to='instagram/')
    children = models.TextField(default='')
    created = models.DateTimeField()

    def children_list(self):
        return self.children.split('\n')

    def __str__(self):
        return self.media_id + ' ' + self.type + ' ' + self.caption



from datetime import timedelta

from instagram_profile.client import convert_auth_code_to_long_lived_token, refresh_long_lived_access_token
from instagram_profile.models import Profile


def update_access_token(profile: Profile, auth_code: str=None, days_buffer=30):
    # If we have an auth_code use it to get a long lived access token and store that in the profile object
    if auth_code:
        result = convert_auth_code_to_long_lived_token(auth_code)
        if result.access_token:
            profile.access_token = result.access_token
            profile.expires_at = result.expires_at
            profile.save()
        else:
            return result.error

    # If we have a long lived access token refresh it if it will expiry soon
    elif profile.access_token and profile.expires_in <= timedelta(days=days_buffer):
        result = refresh_long_lived_access_token(profile.access_token)
        if result.access_token:
            profile.access_token = result.access_token
            profile.expires_at = result.expires_at
            profile.save()
        else:
            return result.error

    # Otherwise, do not do anything as we already have access or can't get access



# Generated by Django 3.1.3 on 2021-01-04 13:44

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Profile',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('username', models.CharField(max_length=256)),
                ('access_token', models.CharField(blank=True, help_text='The long lived access token', max_length=256)),
                ('expires_at', models.DateTimeField(blank=True, help_text='The expiry of the stored long lived access token', null=True)),
                ('modified', models.DateTimeField(auto_now=True)),
            ],
            options={
                'db_table': 'instagram_profile',
                'ordering': ['username'],
            },
        ),
        migrations.CreateModel(
            name='Post',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('media_id', models.CharField(max_length=100)),
                ('type', models.CharField(choices=[('IMAGE', 'Image'), ('CAROUSEL_ALBUM', 'Album'), ('VIDEO', 'Video')], default='IMAGE', max_length=20)),
                ('permalink', models.CharField(max_length=256)),
                ('caption', models.TextField(default='')),
                ('thumbnail', models.ImageField(default='', max_length=256, upload_to='instagram/')),
                ('children', models.TextField(default='')),
                ('created', models.DateTimeField()),
                ('profile', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='instagram_profile.profile')),
            ],
            options={
                'db_table': 'instagram_posts',
                'ordering': ['-created'],
            },
        ),
    ]



from django.conf import settings


INSTAGRAM_AUTH_URL = 'https://api.instagram.com/oauth/authorize'
INSTAGRAM_ACCESS_TOKEN_URL = 'https://api.instagram.com/oauth/access_token'
INSTAGRAM_LONG_LIVED_ACCESS_TOKEN_URL = 'https://graph.instagram.com/access_token'
INSTAGRAM_REFRESH_ACCESS_TOKEN_URL = 'https://graph.instagram.com/refresh_access_token'
INSTAGRAM_MEDIA_URL = 'https://graph.instagram.com'
INSTAGRAM_ACCOUNT = None
INSTAGRAM_APP_ID = settings.INSTAGRAM_PROFILE['app_id']
INSTAGRAM_SECRET = settings.INSTAGRAM_PROFILE['secret']
INSTAGRAM_REDIRECT_URL = settings.INSTAGRAM_PROFILE['redirect_url']



from django.contrib import admin
from django.contrib import messages
from django.http import HttpResponseRedirect
from django.shortcuts import get_object_or_404
from django.urls import reverse, re_path
from django.utils import timezone
from django.utils.safestring import mark_safe

from .client import get_authorize_url
from .models import Post, Profile
from .profiles import update_access_token
from .services import sync_instagram


class ProfileAdmin(admin.ModelAdmin):
    list_display = ('username', 'expires_at', 'expires_in')
    search_fields = ('username',)

    def get_urls(self):
        urls = super().get_urls()
        my_urls = [
            re_path('^authorized/?$', self.admin_site.admin_view(self.authorize_callback), name='instagram_authorize_callback'),
            re_path('^(\d+)/authorize/?$', self.admin_site.admin_view(self.authorize_profile), name='instagram_authorize'),
            re_path('^(\d+)/sync/?$', self.admin_site.admin_view(self.sync), name='instagram_sync'),
        ]
        return my_urls + urls

    def authorize_profile(self, request, object_id):
        profile = get_object_or_404(Profile, id=object_id)
        if profile.access_valid:
            messages.warning(request, 'This profile is already authorized')
            return HttpResponseRedirect(reverse('admin:instagram_profile_profile_change', args=[profile.id]))
        else:
            return HttpResponseRedirect(get_authorize_url(object_id))

    def authorize_callback(self, request):
        # If there is a code, this must be the auth callback redirect from Instagram
        auth_code = request.GET.get('code')
        if auth_code:
            profile = get_object_or_404(Profile, id=request.GET.get('state'))

            err = update_access_token(profile, auth_code=auth_code)
            if err:
                messages.error(request, err)
                return HttpResponseRedirect(reverse('admin:instagram_profile_profile_change', args=[profile.id]))

            messages.success(request, f'Successfully authorized for next {profile.expires_in.days} days and will be automatically extended as needed.')
            return HttpResponseRedirect(reverse('admin:instagram_profile_profile_change', args=[profile.id]))

        # Otherwise redirect to instagram to get an auth code
        else:
            messages.warning(request, 'No code available.  Try again')
            return HttpResponseRedirect(reverse('admin:instagram_profile_profile_changelist'))

    def sync(self, request, object_id):
        profile = get_object_or_404(Profile, id=object_id)
        redirect_to = reverse('admin:instagram_profile_profile_change', args=[profile.id])
        if profile.access_valid:
            result = sync_instagram(profile)
            status = messages.SUCCESS if result['status'] else messages.ERROR
            messages.add_message(request, status, result['message'])
            return HttpResponseRedirect(redirect_to)
        else:
            messages.error(request, 'This profile has not been authorized.')
            return HttpResponseRedirect(redirect_to)


class PostAdmin(admin.ModelAdmin):
    list_display = ('media_id', 'created', 'caption', 'type', 'permalink_display')
    list_filter = ('type', 'created', 'profile')
    search_fields = ('media_id', 'caption')
    date_hierarchy = 'created'

    def get_urls(self):
        urls = super().get_urls()
        my_urls = [
            re_path('^sync/?$', self.admin_site.admin_view(self.sync), name='instagram_sync')
        ]
        return my_urls + urls

    def sync(self, request):
        for profile in Profile.objects.exclude(access_token=''):
            if profile.expires_at > timezone.now():
                result = sync_instagram(profile)
                status = messages.SUCCESS if result['status'] else messages.ERROR
                messages.add_message(request, status, result['message'])
            else:
                messages.warning(request, f'Authorization expired for {profile}.')

        return HttpResponseRedirect(reverse('admin:instagram_profile_post_changelist'))

    def permalink_display(self, obj):
        if obj.permalink:
            return mark_safe('<a href="{}" target="_blank">{}'.format(
                obj.permalink,
                obj.permalink.replace('https://www.instagram.com', '')
            ))
    permalink_display.short_description = 'Permalink'


admin.site.register(Profile, ProfileAdmin)
admin.site.register(Post, PostAdmin)



import os
from shutil import copyfileobj

import requests
from PIL import Image
from django.conf import settings
from django.core.files.base import ContentFile
from requests.exceptions import HTTPError

from . import client
from .models import Post, Profile
from .profiles import update_access_token


def sync_instagram(profile: Profile):
    # Make sure the long lived token is not about to expire
    update_access_token(profile)

    try:
        posts = client.get_media_feed(profile.access_token)
    except HTTPError as http_err:
        return {
            'status': False,
            'message': f'HTTP error occurred: {http_err}'
        }
    except Exception as err:
        return {
            'status': False,
            'message': f'API error occurred: {err}'
        }

    created = get_last_synced_date(profile)
    count = 0
    for post in posts:
        if created is None or post['created'] > created:

            # Download the raw bytes of the image.  Abort if something goes wrong.
            thumbnail_bytes = fetch_image(post['thumbnail'])
            if not thumbnail_bytes:
                continue

            # Organise children
            children = []
            if post['type'] == Post.TYPE_ALBUM:
                for item in post['children']:
                    child_thumbnail = download_image(item['thumbnail'], item['media_id'])
                    resize_image(child_thumbnail)
                    children.append(child_thumbnail)

            # Create post
            new_post = Post(
                profile=profile,
                media_id=post['media_id'],
                caption=post['caption'],
                type=post['type'],
                permalink=post['permalink'],
                created=post['created'],
                children='\n'.join(children),
            )

            # Save the original thumbnail image into the model
            # Can use libraries such as imagekit to resize this image when needed
            name = get_file_name(post['thumbnail'], post['media_id'])
            new_post.thumbnail.save(name, ContentFile(thumbnail_bytes))
            new_post.save()
            count += 1

    return {
        'status': True,
        'message': f'Successfully synced {count} new post{f"" if count == 1 else "s"} from {profile}.',
        'count': count,
    }


def get_last_synced_date(profile):
    try:
        last_db_post = Post.objects.filter(profile=profile).latest('created')
        return last_db_post.created
    except Post.DoesNotExist:
        pass


def get_file_name(url, name):
    ext = url.split('?')[0].split('.')[-1]
    return name + '.' + ext


def fetch_image(url):
    r = requests.get(url, stream=True)
    if r.status_code == 200:
        return r.content


def download_image(url, name):
    res = requests.get(url, stream=True)
    if res.status_code == 200:
        base_name = get_file_name(url, name)
        full_name = os.path.join(settings.MEDIA_ROOT, 'instagram/', base_name)
        if not os.path.exists(full_name):
            with open(full_name, 'wb') as out_file:
                copyfileobj(res.raw, out_file)
        return base_name


def resize_image(base_name):
    full_name = os.path.join(settings.MEDIA_ROOT, 'instagram/', base_name)
    img = Image.open(full_name)
    img.thumbnail((624, 800), Image.ANTIALIAS)
    img.save(full_name, progressive=True, quality=100)
